# Generated by Django 3.2.9 on 2021-11-23 12:32

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='MetaDeviceType',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=100)),
                ('vendor', models.CharField(max_length=50)),
                ('sha', models.CharField(max_length=40)),
                ('download_url', models.URLField()),
            ],
        ),
    ]



"""
from rest_framework.serializers import ModelSerializer

from netbox_devicetype_importer.models import MyModel1


class MyModel1Serializer(ModelSerializer):

    class Meta:
        model = MyModel1
        fields = '__all__'
"""



# Generated by Django 3.2.9 on 2021-12-09 12:17

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_devicetype_importer', '0001_netbox_devicetype_importer'),
    ]

    operations = [
        migrations.AddField(
            model_name='metadevicetype',
            name='is_new',
            field=models.BooleanField(default=True),
        ),
        migrations.AlterField(
            model_name='metadevicetype',
            name='download_url',
            field=models.URLField(blank=True, null=True),
        ),
    ]



from extras.plugins import PluginConfig
from .version import __version__


class NetboxdevicetypeimporterConfig(PluginConfig):
    name = 'netbox_devicetype_importer'
    verbose_name = 'DeviceType Importer'
    description = 'Import DeviceType from github repo'
    version = __version__
    author = 'Nikolay Yuzefovich'
    author_email = 'mgk.kolek@gmail.com'
    required_settings = []
    min_version = '3.0.0'
    max_version = '3.1.99'
    default_settings = {
        'repo_owner': 'netbox-community',
        'repo': 'devicetype-library',
        'github_token': '',
        'use_gql': True,
    }


config = NetboxdevicetypeimporterConfig # noqa



import requests

from jinja2 import Template


class GQLError(Exception):
    default_message = None

    def __init__(self, message=None):
        if message is None:
            self.message = self.default_message
        else:
            self.message = message
        super().__init__(message)


class GitHubAPI():
    def __init__(self, url=None, token=None, owner=None, repo=None):
        self.session = requests.session()
        self.session.headers.update({'Accept': 'application/vnd.github.v3+json'})
        if token:
            self.session.headers.update({'Authorization': f'token {token}'})
        self.dt_dir = 'device-types'
        self.url = f'https://api.github.com/repos/{owner}/{repo}/contents/'

    def get_vendors(self):
        result = {}
        url = f'{self.url}{self.dt_dir}'
        response = self.session.get(url)
        if response.ok:
            for vendor in response.json():
                result[vendor['name']] = vendor['path']
        return result

    def get_models(self, vendor):
        result = {}
        url = f'{self.url}{self.dt_dir}/{vendor}'
        response = self.session.get(url)
        if response.ok:
            for model in response.json():
                result[model['name']] = {
                    'path': model['path'],
                    'sha': model['sha'],
                    'download_url': model['download_url']
                }
        return result

    def get_tree(self):
        '''
        {'cisco': {
            '2950.yaml': {'path': '', 'sha': '', 'download_url': ''}
            }
        }
        '''
        result = {}
        vendors = self.get_vendors()
        for vendor in vendors:
            models = self.get_models(vendor)
            result[vendor] = models
        return result

    def get_files(self, data):
        return {}


class GitHubGQLAPI():
    tree_query = """
{
  repository(owner: "{{ owner }}", name: "{{ repo }}") {
    object(expression: "master:{{ path }}") {
      ... on Tree {
        entries {
          name
          type
          object {
            ... on Blob {
              oid
            }
            ... on Tree {
              entries {
                name
                type
                object {
                  ... on Blob {
                    oid
                  }
                }
              }
            }
          }
        }
      }
    }
  }
}
"""
    files_query = """
{
    repository(owner: "{{ owner }}", name: "{{ repo }}") {
        {% for sha, path in data.items() %}
        sha_{{ sha }}: object(expression: "master:{{ root_path }}/{{ path }}") {
            ... on Blob {
                text
            }
        }
        {% endfor %}
    }
}
"""

    def __init__(self, url='https://api.github.com/graphql', token=None, owner=None, repo=None):
        self.session = requests.session()
        self.session.headers.update({'Authorization': f'token {token}'})
        self.path = 'device-types'
        self.url = url
        self.token = token
        self.owner = owner
        self.repo = repo

    def get_query(self, query):
        result = {}
        response = self.session.post(url=self.url, json={'query': query})
        if response.ok:
            result = response.json()
            err = result.get('errors')
            if err:
                # fix that
                raise GQLError(message=err[0].get('message'))
            return result
        else:
            err = response.json()
            raise GQLError(err.get('message'))
        return result

    def get_tree(self):
        result = {}
        template = Template(self.tree_query)
        query = template.render(owner=self.owner, repo=self.repo, path=self.path)
        data = self.get_query(query)
        if not data:
            return result
        for vendor in data['data']['repository']['object']['entries']:
            result[vendor['name']] = {}
            for model in vendor['object']['entries']:
                result[vendor['name']].update({model['name']: {'sha': model['object']['oid']}})
        return result

    def get_files(self, query_data):
        '''
        data = {'sha': 'venodor/model'}
        result = {'sha': 'yaml_text'}
        '''
        result = {}
        if not query_data:
            return result
        template = Template(self.files_query)
        query = template.render(owner=self.owner, repo=self.repo, data=query_data, root_path=self.path)
        data = self.get_query(query)
        for k, v in data['data']['repository'].items():
            result[k.replace('sha_', '')] = v['text']
        return result



from collections import OrderedDict
from urllib.parse import urlencode


from django.conf import settings
from django.db import transaction
from django.contrib import messages
from django.core.exceptions import ObjectDoesNotExist
from django.views.generic import View
from django.http import HttpResponseForbidden
from django.shortcuts import redirect, reverse
from django.utils.text import slugify

from netbox.views import generic
from dcim.models import Manufacturer, DeviceType
from dcim import forms
from utilities.views import ContentTypePermissionRequiredMixin
from utilities.forms import ImportForm, restrict_form_fields
from utilities.exceptions import AbortTransaction, PermissionsViolation

from .models import MetaDeviceType
from .tables import MetaDeviceTypeTable
from .filters import MetaDeviceTypeFilterSet
from .forms import MetaDeviceTypeFilterForm
from .utilities import GitHubAPI, GitHubGQLAPI, GQLError


class MetaDeviceTypeListView(generic.ObjectListView):
    queryset = MetaDeviceType.objects.all()
    filterset = MetaDeviceTypeFilterSet
    filterset_form = MetaDeviceTypeFilterForm
    table = MetaDeviceTypeTable
    action_buttons = ()
    template_name = 'netbox_devicetype_importer/metadevicetype_list.html'


class MetaDeviceTypeLoadView(ContentTypePermissionRequiredMixin, View):
    def get_required_permission(self):
        return 'netbox_devicetype_importer.add_metadevicetype'

    def post(self, request):
        loaded = 0
        created = 0
        updated = 0
        # deleted = 0
        if not request.user.has_perm('netbox_devicetype_importer.add_metadevicetype'):
            return HttpResponseForbidden()
        plugin_settings = settings.PLUGINS_CONFIG.get('netbox_devicetype_importer', {})
        token = plugin_settings.get('github_token')
        use_gql = plugin_settings.get('use_gql')
        repo = plugin_settings.get('repo')
        owner = plugin_settings.get('repo_owner')
        if token and use_gql:
            gh_api = GitHubGQLAPI(token=token, owner=owner, repo=repo)
        else:
            # gql only
            # gh_api = GitHubAPI(token=token, owner=owner, repo=repo)
            gh_api = GitHubGQLAPI(token=token, owner=owner, repo=repo)
        try:
            models = gh_api.get_tree()
        except GQLError as e:
            messages.error(request, message=f'GraphQL API Error: {e.message}')
            return redirect('plugins:netbox_devicetype_importer:metadevicetype_list')

        for vendor, models in models.items():
            for model, model_data in models.items():
                loaded += 1
                try:
                    metadevietype = MetaDeviceType.objects.get(vendor=vendor, name=model)
                    if metadevietype.sha != model_data['sha']:
                        metadevietype.is_new = True
                        # catch save exception
                        metadevietype.save()
                        updated += 1
                    else:
                        metadevietype.is_new = False
                        metadevietype.save()
                    continue
                except ObjectDoesNotExist:
                    # its new
                    metadevietype = MetaDeviceType.objects.create(
                        vendor=vendor,
                        name=model,
                        sha=model_data['sha']
                    )
                    created += 1
        messages.success(request, f'Loaded: {loaded}, Created: {created}, Updated: {updated}')
        return redirect('plugins:netbox_devicetype_importer:metadevicetype_list')


class MetaDeviceTypeImportView(ContentTypePermissionRequiredMixin, View):
    queryset = MetaDeviceType.objects.all()
    filterset = MetaDeviceTypeFilterSet
    filterset_form = MetaDeviceTypeFilterForm

    related_object_forms = OrderedDict((
        ('console-ports', forms.ConsolePortTemplateImportForm),
        ('console-server-ports', forms.ConsoleServerPortTemplateImportForm),
        ('power-ports', forms.PowerPortTemplateImportForm),
        ('power-outlets', forms.PowerOutletTemplateImportForm),
        ('interfaces', forms.InterfaceTemplateImportForm),
        ('rear-ports', forms.RearPortTemplateImportForm),
        ('front-ports', forms.FrontPortTemplateImportForm),
        ('device-bays', forms.DeviceBayTemplateImportForm),
    ))

    def get_required_permission(self):
        return 'netbox_devicetype_importer.add_metadevicetype'

    def post(self, request):
        vendor_count = 0
        errored = 0
        imported_dt = []
        model = self.queryset.model

        if request.POST.get('_all'):
            if self.filterset is not None:
                pk_list = [obj.pk for obj in self.filterset(request.GET, model.objects.only('pk')).qs]
            else:
                pk_list = model.objects.values_list('pk', flat=True)
        else:
            pk_list = [int(pk) for pk in request.POST.getlist('pk')]

        plugin_settings = settings.PLUGINS_CONFIG.get('netbox_devicetype_importer', {})
        token = plugin_settings.get('github_token')
        use_gql = plugin_settings.get('use_gql')
        repo = plugin_settings.get('repo')
        owner = plugin_settings.get('repo_owner')

        if token and use_gql:
            gh_api = GitHubGQLAPI(token=token, owner=owner, repo=repo)
        else:
            # GraphQL only
            # gh_api = GitHubAPI(token=token, owner=owner, repo=repo)
            gh_api = GitHubGQLAPI(token=token, owner=owner, repo=repo)

        query_data = {}
        # check already imported mdt
        already_imported_mdt = model.objects.filter(pk__in=pk_list, is_imported=True)
        if already_imported_mdt.exists():
            for _mdt in already_imported_mdt:
                if DeviceType.objects.filter(pk=_mdt.imported_dt).exists() is False:
                    _mdt.imported_dt = None
                    _mdt.save()
        vendors_for_cre = set(model.objects.filter(pk__in=pk_list).values_list('vendor', flat=True))
        for vendor, name, sha in model.objects.filter(pk__in=pk_list, is_imported=False).values_list('vendor', 'name', 'sha'):
            query_data[sha] = f'{vendor}/{name}'
        if not query_data:
            messages.warning(request, message='Nothing to import')
            return redirect('plugins:netbox_devicetype_importer:metadevicetype_list')
        try:
            dt_files = gh_api.get_files(query_data)
        except GQLError as e:
            dt_files = {}
            messages.error(request, message=f'GraphQL API Error: {e.message}')
            return redirect('plugins:netbox_devicetype_importer:metadevicetype_list')
        # cre manufacturers
        for vendor in vendors_for_cre:
            manu, _ = Manufacturer.objects.get_or_create(name=vendor, slug=slugify(vendor))
            if _:
                vendor_count += 1

        for sha, yaml_text in dt_files.items():
            form = ImportForm(data={'data': yaml_text, 'format': 'yaml'})
            if form.is_valid():
                data = form.cleaned_data['data']
                model_form = forms.DeviceTypeImportForm(data)
                # is it nessescary?
                restrict_form_fields(model_form, request.user)

                for field_name, field in model_form.fields.items():
                    if field_name not in data and hasattr(field, 'initial'):
                        model_form.data[field_name] = field.initial

                if model_form.is_valid():
                    try:
                        with transaction.atomic():
                            obj = model_form.save()

                            for field_name, related_object_form in self.related_object_forms.items():
                                related_obj_pks = []
                                for i, rel_obj_data in enumerate(data.get(field_name, list())):
                                    f = related_object_form(obj, rel_obj_data)
                                    for subfield_name, field in f.fields.items():
                                        if subfield_name not in rel_obj_data and hasattr(field, 'initial'):
                                            f.data[subfield_name] = field.initial
                                    if f.is_valid():
                                        related_obj = f.save()
                                        related_obj_pks.append(related_obj.pk)
                                    else:
                                        for subfield_name, errors in f.errors.items():
                                            for err in errors:
                                                err_msg = "{}[{}] {}: {}".format(field_name, i, subfield_name, err)
                                                model_form.add_error(None, err_msg)
                                        raise AbortTransaction()
                    except AbortTransaction:
                        # log ths
                        pass
                    except PermissionsViolation:
                        errored += 1
                        continue
                if model_form.errors:
                    errored += 1
                else:
                    imported_dt.append(obj.pk)
                    metadt = MetaDeviceType.objects.get(sha=sha)
                    metadt.imported_dt = obj.pk
                    metadt.save()
            else:
                errored += 1
        # msg
        if imported_dt:
            messages.success(request, f'Imported: {imported_dt.__len__()}')
            if errored:
                messages.error(request, f'Failed: {errored}')
            qparams = urlencode({'id': imported_dt}, doseq=True)
            return redirect(reverse('dcim:devicetype_list') + '?' + qparams)
        else:
            messages.error(request, 'Can not import Device Types')
            return redirect('plugins:netbox_devicetype_importer:metadevicetype_list')



from django.urls import path


from .views import MetaDeviceTypeListView, MetaDeviceTypeLoadView, MetaDeviceTypeImportView

urlpatterns = [
    path('meta-device-types/', MetaDeviceTypeListView.as_view(), name='metadevicetype_list'),
    path('meta-device-types/load/', MetaDeviceTypeLoadView.as_view(), name='metadevicetype_load'),
    path('meta-device-types/import/', MetaDeviceTypeImportView.as_view(), name='metadevicetype_import')
]



# Generated by Django 3.2.9 on 2021-12-10 15:35

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('netbox_devicetype_importer', '0003_netbox_devicetype_importer'),
    ]

    operations = [
        migrations.AddField(
            model_name='metadevicetype',
            name='is_imported',
            field=models.BooleanField(default=False),
        ),
    ]



from extras.plugins import PluginMenuItem


menu_items = (
    PluginMenuItem(
        link='plugins:netbox_devicetype_importer:metadevicetype_list',
        link_text='DeviceType Import',
        permissions=['netbox_devicetype_importer.view_metadevicetype'],
        buttons=(
        )
    ),
)



from django.db import models

from utilities.querysets import RestrictedQuerySet


class MetaDeviceType(models.Model):
    name = models.CharField(max_length=100)
    vendor = models.CharField(max_length=50)
    sha = models.CharField(max_length=40)
    download_url = models.URLField(null=True, blank=True)
    is_new = models.BooleanField(default=True)
    imported_dt = models.IntegerField(null=True, blank=True)
    is_imported = models.BooleanField(default=False)

    objects = RestrictedQuerySet.as_manager()

    def __str__(self):
        return self.name.split('.')[0]

    def save(self, *args, **kwargs):
        if self.imported_dt:
            self.is_imported = True
            self.is_new = False
        else:
            self.is_imported = False
        super(MetaDeviceType, self).save(*args, **kwargs)



"""
from rest_framework.viewsets import ModelViewSet
from netbox_devicetype_importer.models import MyModel1
from .serializers import MyModel1Serializer


class MyModel1ViewSet(ModelViewSet):
    queryset = MyModel1.objects.all()
    serializer_class = MyModel1Serializer
"""



import django_filters

from django.db.models import Q

from .models import MetaDeviceType


class MetaDeviceTypeFilterSet(django_filters.FilterSet):
    q = django_filters.CharFilter(
        method='search',
        label='Search',
    )

    name = django_filters.CharFilter(
        method='by_model',
        label='Model',
    )

    vendor = django_filters.CharFilter(
        method='by_vendor',
        label='Vendor',
    )

    class Meta:
        model = MetaDeviceType
        fields = ['name', 'vendor']

    def by_model(self, queryset, name, value):
        if not value.strip():
            return queryset

        return queryset.filter(Q(name__icontains=value))

    def by_vendor(self, queryset, name, value):
        if not value.strip():
            return queryset
        if ',' in value:
            q = Q()
            for _ in value.split(','):
                if _:
                    q |= Q(vendor__icontains=_)
            return queryset.filter(q)
        return queryset.filter(Q(vendor__icontains=value))

    def search(self, queryset, name, value):
        if not value.strip():
            return queryset

        qs_filter = (
            Q(name__icontains=value) | Q(vendor__icontains=value)
        )
        return queryset.filter(qs_filter)



from django import forms

from utilities.forms import BootstrapMixin

from .models import MetaDeviceType


class MetaDeviceTypeFilterForm(BootstrapMixin, forms.Form):
    q = forms.CharField(
        required=False,
        label='Search'
    )

    name = forms.CharField(
        required=False,
        label='Model'
    )
    vendor = forms.CharField(
        required=False,
        label='Vendor'
    )

    class Meta:
        model = MetaDeviceType
        fields = ['q', 'name', 'vendor']



from utilities.tables import BaseTable, ToggleColumn

from .models import MetaDeviceType


class MetaDeviceTypeTable(BaseTable):
    pk = ToggleColumn()
    id = None

    def render_name(self, value):
        return '{}'.format(value.split('.')[0])

    class Meta(BaseTable.Meta):
        model = MetaDeviceType
        fields = ('pk', 'name', 'vendor', 'is_new', 'is_imported')
        default_columns = ('pk', 'name', 'vendor', 'is_imported')



import codecs
import os.path

from setuptools import find_packages, setup


with open("README.md", "r") as fh:
    long_description = fh.read()


def read(rel_path):
    here = os.path.abspath(os.path.dirname(__file__))
    with codecs.open(os.path.join(here, rel_path), 'r') as fp:
        return fp.read()


def get_version(rel_path):
    for line in read(rel_path).splitlines():
        if line.startswith('__version__'):
            delim = '"' if '"' in line else "'"
            return line.split(delim)[1]
    else:
        raise RuntimeError("Unable to find version string.")


setup(
    name='netbox-devicetype-importer',
    version=get_version('netbox_devicetype_importer/version.py'),
    description='Import DeviceTypes from github repo',
    long_description=long_description,
    long_description_content_type="text/markdown",
    url='https://github.com/k01ek/netbox-devicetype-importer',
    author='Nikolay Yuzefovich',
    author_email='mgk.kolek@gmail.com',
    install_requires=[],
    packages=find_packages(),
    include_package_data=True,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Framework :: Django',
        'Programming Language :: Python :: 3',
    ]
)
